// 处理Java 类型的项目构建
var _ = require('lodash');
const execa = require('execa');
const replace = require('replace-in-file');
const fs = require('fs');
var sleep = require('sleep');
var pomParser = require("pom-parser");
var databaseHelper = require("../tools/databaseHelper");
var moment = require('moment');
var config = require("../config/common");

/**
 * 构建Java项目
 * @param {*} projectObj 项目信息
 * @param {*} envObj 环境信息
 * @param {*} typeObj 构建源数据
 * @param {*} configList 配置信息
 * @param {*} branch 分支
 * @param {*} version 版本号
 */
const buildJava = async (projectObj, envObj, typeObj, configList, branch, version, is_deploy, deploy_list) => {
    //start to buid 
    // 获取项目源数据
    var projectMetaData = JSON.parse(projectObj.meta_data);
    console.log(JSON.stringify(projectMetaData));

    // 下载到的地址
    var source_git_url = projectObj.source_git_url;
    // 项目名
    var project_name = envObj.env_id + "-" + projectObj.name;
    // 项目文件夹
    var project_dir = __dirname + '/../tmp/' + project_name;
    // gitlab操作令牌
    var token = envObj.target_git_token;
    // 目标gitlab地址
    var target_git_url = envObj.target_git_url;
    // 目标gitlab用户名
    var target_git_name = envObj.target_git_name;
    // 仓库分组地址
    var registry_group = envObj.registry_group;

    var registry_url = envObj.registry_url;

    // 删除原项目
    let { stdout1 } = await execa('rm', ['-rf', project_dir]);
    console.log(JSON.stringify("删除文件夹" + project_dir));
    //下载git项目
    if (_.isEmpty(branch)) {
        branch = "master";
    }
    let { stdout2 } = await execa(`git`, ['clone', '-b', branch, source_git_url, project_name], { cwd: __dirname + '/../tmp/' })
    console.log("下载项目")

    // 删除.git 等文件
    let { stdout3 } = await execa('rm', ['-rf', project_dir + '/.git']);
    console.log("删除.git");

    //gitlab 创建项目
    let stdout4 = await execa('curl', [
        '--header',
        `PRIVATE-TOKEN:${token}`,
        '-X',
        'DELETE',
        `${target_git_url}/api/v4/projects/${target_git_name}%2f${project_name}`
    ]);
    console.log(stdout4);

    // sleep
    sleep.sleep(3);

    let stdout5 = await execa('curl', [
        '--header',
        `PRIVATE-TOKEN:${token}`,
        '-X',
        'POST',
        `${target_git_url}/api/v4/projects?name=${project_name}`
    ]);

    console.log(stdout5);
    var git_info = JSON.parse(stdout5.stdout);
    //获取项目远程地址
    var git_remote = git_info.http_url_to_repo;
    var project_id = git_info.id;
    console.log("git项目的远程地址为:");
    console.log(git_remote);

    //配置文件替换
    if (!_.isEmpty(configList)) {
        configList.forEach((config) => {
            (async () => {
                let configType = config.type;
                switch (configType) {
                    case 1:
                        // 替换模式
                        var key_value = JSON.parse(config.content);
                        await replaceFileByMap(key_value, `${project_dir}/${config.path}`, config.content)
                        break;
                    case 2:
                        // 覆盖模式
                        fs.writeFileSync(`${project_dir}/${config.path}`, config.content);
                        break;
                    default:
                        console.warn("Unknow Config type");
                }
            })();
        })
    }

    // 先看看build语句
    // 默认跳过测试
    var build_script = _.get(projectMetaData, "build_script", "    - mvn clean install package -Dmaven.test.skip=true");
    // 识别要扫描的模块 
    var module_list = _.get(projectMetaData, "module_list", ["."]);
    // 初始化变量
    var variables_str = "";
    var cache = "";
    var docker_build_str = "";
    var docker_tag_str = "";
    var docker_push_str = "";
    var deploy_str = "";

    // todo 处理docker登录问题
    var username = _.get(envObj, 'registry_username');
    var password = _.get(envObj, 'registry_password');
    // docker login ${IMAGE_HOST} -u yzy -p 123qweQWE@ 

    if (!_.isEmpty(username) && !_.isEmpty(password)) {
        docker_build_str += "    - docker login ${IMAGE_HOST} -u " + username + " -p " + password + "\n";
        // docker_build_str += "    - docker login 10.100.9.22:2080 -u " + username + " -p " + password + "\n";
    }

    let image_names = "";
    // 关键替换
    for (let v of module_list) {
        let opt = "";
        var module_name = "";

        //构建项目的服务名
        var deploy_name = "";

        if (_.isString(v)) {
            module_name = v;
            opt = { filePath: project_dir + "/" + v + "/pom.xml" }
        } else if (_.isObject(v)) {
            module_name = v.name;
            opt = { filePath: project_dir + "/" + v.name + "/pom.xml" }
        } else {
            throw new Error("错误的类型")
        }

        deploy_name = envObj.name + "-" + module_name;
        if (_.isObject(v) && _.has(v, `deploy.${envObj.name}`)) {
            deploy_name = _.get(v, `deploy.${envObj.name}`);
        }

        var pom_object = await pomParserPromise(opt);
        // console.log(JSON.stringify(pom_object));
        // 这里默认情况下使用通配符 当配置时覆盖
        var exec_file = pom_object.project.name + "*";
        if (_.isObject(v) && _.has(v, 'exec-file')) {
            exec_file = _.get(v, "exec-file");
        }
        var image_name = module_name.toLocaleLowerCase();
        var service_name = _.replace(pom_object.project.name, '-', '_');
        image_names += `${registry_group}/${project_name}/${image_name}` + "|";
        var dockerfile_name = "dockerfile_" + service_name;
        variables_str += "  " + "IMAGE_NAME_" + service_name + ': "${IMAGE_HOST}/' + registry_group + '/${CI_PROJECT_NAME}/' + image_name + '"' + "\n";
        variables_str += "  " + 'IMAGE_VERSION_' + service_name + ': "${IMAGE_HOST}/' + registry_group + '/${CI_PROJECT_NAME}/' + image_name + ':${CI_COMMIT_TAG}"' + "\n";
        variables_str += "  " + 'IMAGE_LATEST_' + service_name + ': "${IMAGE_HOST}/' + registry_group + '/${CI_PROJECT_NAME}/' + image_name + ':latest"' + "\n";
        //缓存地址
        cache += "    - " + module_name + "/target/\n";
        docker_build_str += "    " + "- docker build -t ${IMAGE_VERSION_" + service_name + "} -f " + dockerfile_name + " ." + "\n";
        docker_tag_str += "    " + '- docker tag ${IMAGE_VERSION_' + service_name + '} ${IMAGE_LATEST_' + service_name + '}' + "\n";
        docker_push_str += "    " + '- docker push ${IMAGE_VERSION_' + service_name + '}' + "\n";
        docker_push_str += "    " + '- docker push ${IMAGE_LATEST_' + service_name + '}' + "\n";

        //部署指令
        if (is_deploy && deploy_list.indexOf(module_name) > 0) {
            deploy_str += "    " + `- docker service update --image ${envObj.registry_url}/${registry_group}/${project_name}/${image_name}:${version} ${deploy_name}` + "\n"
        }
        // 这里要写dockerfile

        fs.writeFileSync(`${project_dir}/${dockerfile_name}`, typeObj.docker_file);
        var dockerfile_map = {
            "{service}": module_name,
            "{service-exec}": exec_file
        }
        await replaceFileByMap(dockerfile_map, `${project_dir}/${dockerfile_name}`);
    }
    console.log("部署脚本");
    console.log(deploy_str);

    // 在这里保存数据正好
    var log_data = {
        image_name: image_names,
        p_id: projectObj.p_id,
        env_id: envObj.env_id,
        project_name: projectObj.name,
        branch: branch,
        version: version,
        success: 3,
        create_time: moment().format("YYYY-MM-DD HH:mm:ss")
    }

    var logObject = await databaseHelper.saveJsonToDb(log_data, 'build_log', 'log_id');
    console.log(JSON.stringify(logObject));

    //设置成功回调
    var success_str = "    - \"curl -X PUT http://" + config.host + ":" + config.port + "/build/log/back -H 'Content-Type: application/json' " + `-d '{\\"log_id\\":${logObject.log_id},\\"success\\":1}'"`;
    // 失败回调
    var fail_str = "    - \"curl -X PUT http://" + config.host + ":" + config.port + "/build/log/back -H 'Content-Type: application/json' " + `-d '{\\"log_id\\":${logObject.log_id},\\"success\\":2}'"`;

    // 创建gitlab-ci
    fs.writeFileSync(`${project_dir}/.gitlab-ci.yml`, typeObj.gitlab_ci);
    var gitlab_ci_map = {
        "${registry_url}": registry_url,
        "{variables_str}": variables_str,
        "{cache}": cache,
        "{docker_build_str}": _.trimEnd(docker_build_str, "\n"),
        "{docker_tag_str}": _.trimEnd(docker_tag_str, "\n"),
        "{docker_push_str}": _.trimEnd(docker_push_str, "\n"),
        "{build_str_java}": build_script,
        "{success_str}": success_str,
        "{fail_str}": fail_str,
        "{deploy_str}": _.trimEnd(deploy_str, "\n"),
        "{tag}": envObj.name,
    };
    await replaceFileByMap(gitlab_ci_map, `${project_dir}/.gitlab-ci.yml`);
    // 提供gitlab-ci

    // 创建数据//



    // todo 创建docker-compose
    // 替换docker-compose

    // 提交上传代码
    let { stdout9 } = await execa(`git`, ['init'], { cwd: project_dir })
    // console.log(stdout9);
    let { stdout10 } = await execa(`git`, ['add', '.'], { cwd: project_dir });
    let { stdout12 } = await execa(`git`, ['commit', '-m', '"Init"'], { cwd: project_dir });
    let { stdout13 } = await execa(`git`, ['remote', 'add', 'origin', git_remote], { cwd: project_dir });
    let { stdout14 } = await execa(`git`, ['push', '-u', 'origin', 'master'], { cwd: project_dir });
    // 打上tag
    let stdout15 = await execa('curl', [
        '--header',
        `PRIVATE-TOKEN:${token}`,
        '-X',
        'POST',
        `${target_git_url}/api/v4/projects/${project_id}/repository/tags?tag_name=${version}&ref=master`
    ]);
    console.log(stdout15.stdout);
}

/**
 * 按照key-value的map 替换一组值
 * @param {*} map  替换字符串对
 * @param {*} patch  文件路径
 */
const replaceFileByMap = async (map, patch) => {
    const quote = function (str) {
        return (str + '').replace(/[.?*+^$[\]\\(){}|-]/g, "\\$&");
    };
    if (!_.isEmpty(map)) {
        var from = [];
        var to = [];
        _.forEach(map, function (value, key) {
            (async () => {
                let quoteKey = quote(key);
                // console.log(quoteKey);
                from.push(eval(`/${quoteKey}/g`));
                to.push(value);
            })();
        });
        let options = {
            files: patch,
            from: from,
            to: to,
        };
        let loga = await replace(options);
        console.log(loga);
    }
}


/**
 * promise 版的解析Pom文件
 * @param {*} opt 
 */
var pomParserPromise = (opts) => {
    return new Promise((reslove, reject) => {
        pomParser.parse(opts, function (err, pomResponse) {
            if (err) {
                console.log("ERROR: " + err);
                reject(err);
            }
            reslove(pomResponse.pomObject);
        });
    });
}

module.exports = {
    buildJava,
}